package CTL.Comm;

import CTL.*;
import CTL.Streams.*;
import CTL.Types.*;

import java.io.*;
import java.lang.reflect.*;
import java.net.*;

/** TCP/IP communication handling */
public class TCPCommunicator implements Communicator
{
	/** Local server socket */
	private static ServerSocket ssock = null;
	/** Local IP */
	private static IPaddr host = null;
	/** Local port */
	private static int port = -1;
	/** Socket connection to the communication partner */
	private Socket sock = null;

	/** Change the contained IP address
	 * @param host Host
	 */
	public void setHost (IPaddr host)
	{
		this.host = host;
	}

	/** Constructor for known remote host and port */
	public TCPCommunicator (String host, int port)
	{
		this.host = determineIP(host, port);
	}

	/** Constructor for a yet to be made connection */
	public TCPCommunicator ()
	{
		this.port = 0;
	}

	/** Constructor for known remote port */
	public TCPCommunicator (int port)
	{
		this.port = port;
	}

	/** Listen for new connections */
	public void listen () throws IOException
	{
		ssock = new ServerSocket(port, 25);
		port = ssock.getLocalPort();
		Env.log.log_msg(Logger.DBG1, "Listening on "+ssock.getLocalPort()+".");
	}

	/** Retrieve the PeerID
	 * @return PeerID
	 */
	public PeerID pid ()
	{
		return new PeerID(host, port);
	}

	/** Connect to peer 
	 * @param pid PeerID of remote host
	 */
	public void connect (PeerID pid) 
		throws UnknownHostException, IOException
	{
		sock = new Socket(pid.host().toString(), pid.port());
		this.host = new IPaddr(sock.getLocalAddress());	
		this.port = sock.getLocalPort();
	}

	/** Accept a connection */
	public void accept () throws IOException
	{
		sock = ssock.accept();
		// Now the real IP can be determined from the connection's socket.
		host = new IPaddr(sock.getLocalAddress());
	}

	/** Close the connection */
	public void close () throws IOException
	{
		sock.close();
	}

	/** Send a specific amount of data
	 * @param data Bytes to send
	 * @param len Number of bytes to send
	 */
	public void send (byte[] data, int len) throws IOException
	{
		send(data);
	}
	
	/** Send a specific amount of data
	 * @param data Bytes to send
	 */
	public void send (byte[] data) throws IOException
	{
		OutputStream out = sock.getOutputStream();
		Env.log.log_msg(Logger.INFO, "Writing "+(data.length-36)+" bytes.");
		out.write(data);
		out.flush();
	}

	/** Receive a specific amount of data
	 * @param len Number of bytes to receive
	 * @return Data
	 */
	public byte[] recv (int len) throws IOException, CTLException
	{
		InputStream inp = sock.getInputStream();
		OIStream in = new OIStream(inp);
		byte[] buf = new byte[len];
		inp.read(buf);
		return buf;
	}

	/** Receive a CTL packet (length is determined from header)
	 * @return Data
	 */
	public byte[] recv () throws IOException, CTLException
	{
		try
		{
			InputStream inp = sock.getInputStream();
			OIStream in = new OIStream(inp);
			Header head = in.serialRead(Header.class);
			if (head == null)
				return null;
			byte[] buf = new byte[(int)head.getSize()];
			Env.log.log_msg(Logger.INFO, "Read "+head.getSize()+" bytes.");
			inp.read(buf);
		
			BuffyOut out = new BuffyOut();
			out.o.serialWrite(head);
			if (buf.length>0)
				out.o.write(buf, 0, buf.length);
			out.o.flush();
			return out.getBytes();
		} 
		catch (IllegalAccessException e0) {}
		catch (ClassNotFoundException e1) {}
		catch (InvocationTargetException e2) {}
		catch (InstantiationException e3) {}

		return null;
	}

	/** Open a temporary TCP connection to determine the outgoing IP for
	 * communication with the target IP
	 * @return Determined IP address
	 */
	protected static IPaddr determineIP (String host, int port)
	{
		try
		{
			Socket sock = new Socket(host, port);
			IPaddr ret = new IPaddr(sock.getLocalAddress());
			sock.close();
			return ret;
		}
		catch (IOException e)
		{
			return null;
		}
	}
}
