package CTL;

import CTL.Comm.*;
import CTL.Streams.*;
import CTL.Types.*;

import java.io.*;
import java.lang.reflect.*;
import java.net.*;

/** Incoming connection handling thread */
class ConnectHandler extends Thread
{
	/** Handles a connection to Env.sock
	 * @see Env
	 */
	public void run ()
	{
		// GROUP: Determine ID somehow
		int conn = (Env.grp.myID() == 0 ? 1 : 0);
		
		try
		{
			accept(conn);
		}
		catch (Exception e)
		{
			if (!Env.grp.running())
				return;
			RUtil.except(e);
			try
			{
				Env.comm[conn].close();
			} catch (IOException e2) {}
		}
	}
	
	/** Accept a connection and delegate it to a handling method
	 * @param id Logical ID of the connecting client
	 */
	private void accept (int id) throws IOException, IllegalAccessException,
		InvocationTargetException, CTLException, NoSuchMethodException,
		ClassNotFoundException, InstantiationException
	{
		byte[] msg = Env.comm[id].recv();
		
		if (msg == null)
		{
			if (!Env.grp.running())
				return;
			if (Env.daemon)
				return;
			Env.log.log_msg(Logger.ERR, "No data received.");
		}
			
		BuffyIn in = new BuffyIn(msg);
		Header ohead = null;
		long objID;
		
		ohead = in.i.serialRead(Header.class);
		Env.log.log_msg(Logger.INFO, "Recv header: "+ohead);

		switch (ohead.tag())
		{
			case Remote.OPER:
				Env.log.log_msg(Logger.ERR, "OPER is not yet supported.");
				System.exit(5);
				break;

			case Remote.DAT:
				Env.log.log_msg(Logger.ERR, "DAT is not yet supported.");
				System.exit(5);
				break;
			
			case Remote.RMI:
				objID = in.i.readLong();
				procRMI(in.i, ohead, objID, id);
				break;

			default:
				objID = in.i.readLong();
		
				Env.log.log_msg(Logger.INFO, ".:. Received RMI answer .:.");
				Env.log.log_msg(Logger.INFO, "\tHeader: "+ohead);
				Env.log.log_msg(Logger.INFO, "\tObjectID: "+objID);

				Object obj = Env.map.getObj((int)objID);
				if (obj==null)
				{
					Env.log.log_msg(Logger.ERR, "Object "+objID+" not found.");
					System.exit(10);
				}

				if (!obj.getClass().getName().equals("CTL.rResult"))
				{
					procRMI(in.i, ohead, objID, 1);
					return;
				}

				Except ex = in.i.serialRead(Except.class);
				if (ex.name().length()>0)
					throw new CTLException("Class not found!");

				rResult res = (rResult)obj;	
				for (int i=0;i<res.results();i++)
					res.recv(in.i, i);
				break;
		}
	}

	/** Calculates how long the argument list of a call is
	 * @param head Header of this package
	 * @param fid FunctionID of the invoked method
	 * @return Length of the argument list in bytes
	 */
	private static long argSize (Header head, FID fid)
	{
		return head.getSize()-47-fid.name().length();
	}
	
	/** Process an RMI call
	 * @param in Input stream for this connection
	 * @param head Header of this package
	 * @param objID Requested ObjectID
	 * @param member Local group member ID
	 */
	private static void procRMI (OIStream in, Header head, long objID, 
		int member) throws IOException, IllegalAccessException, 
		NoSuchMethodException, InvocationTargetException, 
		ClassNotFoundException, CTLException, InstantiationException
	{
		FID fid = in.serialRead(FID.class);
		Class targetClass = null;

		try
		{
			Env.log.log_msg(Logger.INFO, "FID: ObjectID: "+objID+
				" - Target class: "+fid._class());
			//Env.log.log_msg(Logger.INFO, Env.map.dump());
			
			if (objID<=0)
				targetClass = Class.forName(fid._class());
			else
			{
				Object req = Env.map.getObj((int)objID);
				if (req==null)
					throw new ClassNotFoundException("No such ObjectID.");
				targetClass = RUtil.classToRI(req.getClass());
			}
		}
		catch (ClassNotFoundException e)
		{
			Env.log.log_msg(Logger.ERR, "Class not found: \""+
				fid._class()+"\" ("+e.getMessage()+")");
		
			// Read junk away
			int[] sz = new int[2];
			sz[0] = (int)argSize(head,fid);
			sz[1] = (int)(argSize(head, fid) >> 32);
			for (int i=0;i<2;i++)
			{
				byte[] junk = new byte[sz[i]];
				int moo = in.read(junk);
			}
			
			rPointer rp = in.serialRead(rPointer.class);
			int tag = in.readInt();
			
			Header head2 = new Header(0, tag, Env.comm[member]);
			Except ex = new Except("Class not found.");
			Remote.answer(Env.comm[member], head2, rp.objID(), ex, null);
			return;
		}
		
		Env.log.log_msg(Logger.DBG, "ObjectID: "+objID+" - Target class: "+
			targetClass.getName());

		Class[] args = new Class[5];
		args[0] = OIStream.class;
		args[1] = Header.class;
		args[2] = Long.TYPE;
		args[3] = FID.class;
		args[4] = Integer.TYPE;
		
		Object[] args2 = new Object[5];
		args2[0] = (Object)in;
		args2[1] = (Object)head;
		args2[2] = new Long(objID);
		args2[3] = (Object)fid;
		Env.log.log_msg(Logger.DBG, ""+member);
		Env.log.log_msg(Logger.DBG, ""+Env.grp);
		args2[4] = new Integer(member);

		Method m = targetClass.getDeclaredMethod("accept", args);
		m.invoke(null, args2);
	}
}
