package CTL;

import CTL.Annotate.*;
import CTL.Comm.*;
import CTL.Types.*;
import ReflWrap.*;

import java.io.*;
import java.lang.reflect.*;
import java.net.*;
import java.util.*;

/** Process group (currently, only groups with 2 members are supported) */
public class Group implements Runnable
{
	private final static int logtype = Logger.INFO;
	
	/** Logical ID of the process running on localhost */
	private int myID;
	/** Number of hosts in this group */
	private int members;
	/** ObjectID */
	private long objID;
	/** Array of GroupInfos for all members */
	protected GroupInfo[] grp;
	/** Whether the group is terminated */
	private boolean terminated = true;

	/** Receive a termination request
	 * @param pid PeerID of the terminating process
	 * @param msg Reason for termination
	 */
	@builtin @sFID(3) public void recvTermination (PeerID pid, String msg)
	{
		Env.log.log_msg(Logger.INFO, "recv TERM: "+pid+" - "+msg);
		System.exit(0);
	}

	/** Receive an exception
	 * @param msg Reason for the exception
	 */
	@builtin @sFID(4) public static void recvException (String msg)
	{
		// TODO: weitermachen bzw evtl mit recvTermination beenden
		throw new RuntimeException(msg);
	}

	// Client which starts a remote service via SSH
	@cdont public Group (String host, int port, int myID, int members,
		int link) throws IOException, CTLException
	{
		Env.log = new Logger("/tmp/client-"+myID+".log", logtype);
		objID = Env.map.regObj(this);
		Env.log.log_msg(Logger.DBG1, "Registered as "+objID+".");
		
		Env.comm = new Communicator[members];

		switch (link)
		{
			case Location.TCP:
				Env.comm[0] = new TCPCommunicator();
				Env.comm[1] = new TCPCommunicator(host, port);
				break;
			default:
				throw new CTLException("Linkage not implemented.");
		}
		
		Env.comm[1].listen();

		updateGroup(myID, members);
		grp[myID] = new GroupInfo(Env.comm[0].pid(), objID);
	}

	// Communication with a daemon
	@cdont public Group (String host, int port, int myID, int members, 
		int link, boolean is_service) throws IOException, 
		ClassNotFoundException, IllegalAccessException, 
		InstantiationException, InvocationTargetException, CTLException
	{
		if (is_service)
		{
			Env.log = new Logger("/tmp/service.log", logtype);
			objID = Env.map.regObj(this);
			Env.log.log_msg(Logger.DBG1, "Registered as "+objID+".");
			
			Env.comm = new Communicator[members];
			
			switch (link)
			{
				case Location.HTTP:
					Env.comm[0] = new HTTPCommunicator();
					Env.comm[1] = new HTTPCommunicator();
					((HTTPCommunicator)Env.comm[0]).setPort(port);
					break;
				case Location.TCP:
					Env.comm[0] = new TCPCommunicator(port);
					Env.comm[1] = new TCPCommunicator(port);
					break;
				default:
					throw new CTLException("Linkage not implemented.");
			}
			
			Env.comm[0].listen();
			updateGroup(myID, members);
			
			if (link == Location.TCP)
			{
				((TCPCommunicator)Env.comm[0]).setHost(new IPaddr(host));
				System.out.println(Env.comm[0].pid().port());
			}
			
			grp[myID] = new GroupInfo(Env.comm[0].pid(), objID);
			Env.daemon = true;
		}
		else
		{
			Env.log = new Logger("/tmp/client-"+myID+".log", logtype);
			objID = Env.map.regObj(this);
			Env.log.log_msg(Logger.DBG1, "Registered as "+objID+".");

			Env.comm = new Communicator[members];
			
			switch (link)
			{
				case Location.HTTP:
					Env.comm[0] = new HTTPCommunicator();
					Env.comm[1] = new HTTPCommunicator();
					break;
				case Location.TCP:
				case Location.DMN:
					Env.comm[0] = new TCPCommunicator();
					Env.comm[1] = new TCPCommunicator();
					break;
				default:
					throw new CTLException("Linkage not implemented.");
			}
				
			Env.comm[0].listen();
			Env.comm[0].connect(new PeerID(host, port));
			readVersion(Env.comm[0]);
			
			updateGroup(myID, members);
			grp[0] = new GroupInfo(new PeerID(host, port), 0);
			grp[1] = new GroupInfo(Env.comm[1].pid(), objID);

			Remote.writeDAT(Env.comm[0], null, new IPaddr(Remote.magic, 0),
				Env.comm[1].pid().port());
		}
	}

	// Remote service, started via SSH
	@cdont public Group (String[] args) throws IOException, CTLException, 
		IllegalAccessException, InvocationTargetException,
		ClassNotFoundException, InstantiationException
	{
		Env.log = new Logger("/tmp/service.log", logtype);
		objID = Env.map.regObj(this);
		Env.log.log_msg(Logger.DBG1, "Registered as "+objID+".");
		
		StringTokenizer token = new StringTokenizer(args[0], ":");
		String proto = token.nextToken();
		IPaddr host = new IPaddr(token.nextToken());
		int port = Integer.parseInt(token.nextToken());
		int myID = Integer.parseInt(token.nextToken());
		int members = Integer.parseInt(token.nextToken());

		Env.log.log_msg(Logger.DBG, 
			proto+"://"+host+":"+port+" - "+myID+"/"+members);

		Env.comm = new Communicator[members];

		if (proto.equals("tcp"))
		{
			Env.comm[0] = new TCPCommunicator();
			Env.comm[1] = new TCPCommunicator();
		}
		
		Env.comm[0].listen();
		Env.comm[0].connect(new PeerID(host, port));
		readVersion(Env.comm[0]);

		updateGroup(myID, members);
		grp[1] = new GroupInfo(Env.comm[0].pid(), objID);

		Env.log.log_msg(Logger.DBG1, "Sending handshake to "+
			host+":"+port);
		HS_send();
	}

	/** Update group information
	 * @param myID Logical ID
	 * @param members Number of group members
	 */
	@dont private void updateGroup (int myID, int members)
	{
		this.myID = myID;
		this.members = members;
		if (members>0)
			grp = new GroupInfo[members];
	}
	
	/** Print information about the group */
	@dont public void printInfo ()
	{
		Env.log.log_msg(Logger.DBG1, "I am "+toString()+" on "+
			grp[myID].toString());
	}

	/** Read version string
	 * @param comm Communicator
	 */
	@dont public void readVersion (Communicator comm) throws IOException,
		CTLException
	{
		byte[] buf = comm.recv(19);
		Env.log.log_msg(Logger.INFO, new String(buf));
	}
	
	/** Write version string
	 * @param comm Communicator
	 */
	@dont public void writeVersion (Communicator comm) throws IOException
	{
		String str = "CTL v0.9 Java      ";
		comm.send(str.getBytes());
	}

	/** Start the communication thread */
	@dont public void run ()
	{
		if (Env.daemon)
		{
			try
			{
				Env.comm[1].accept();
				writeVersion(Env.comm[1]);
				Header head = (Header)Remote.readDAT(Env.comm[1], "HEAD");
				IPaddr magic = new IPaddr(Remote.magic, 0);
				if (!head.pid().host().equals(magic))
				{
					Env.log.log_msg(Logger.ERR, 
						"Magic numbers do not match.");
					throw new RuntimeException("Magic numbers do not match.");
				}
			} 
			catch (Exception e) 
			{
				RUtil.except(e);
			}
		}
		recv();
	}

	/** Spawns a connection handler as long as the group is active */
	@dont protected void recv ()
	{
		terminated = false;
		while (!terminated)
			new ConnectHandler().run();
	}

	/** Receive a handshake */
	@dont protected void HS_recv () throws IOException, 
		ClassNotFoundException, InstantiationException, 
		InvocationTargetException, IllegalAccessException,
		CTLException
	{
		BuffyIn in = new BuffyIn(Env.comm[1].recv());
		Header head = in.i.serialRead(Header.class);
		
		Remote.writeHS(grp[0], Env.comm[1]);
		grp[1] = Remote.readHS(Env.comm[1]);
		Env.log.log_msg(Logger.INFO, "Handshake received.");
	}
		
	/** Send a handshake */
	@dont protected void HS_send () throws IOException, 
		IllegalAccessException, InvocationTargetException,
		ClassNotFoundException, InstantiationException, 
		CTLException
	{
		Header head = new Header(0, Remote.DAT, new IPaddr(Remote.magic, 0),
			Env.comm[0].pid().port(), myID());
		BuffyOut out = new BuffyOut();
		out.o.serialWrite(head);
		out.o.flush();
		Env.comm[0].send(out.getBytes());
		
		Env.log.log_msg(Logger.INFO, "Sending handshake.");
		grp[0] = Remote.readHS(Env.comm[0]);
		Remote.writeHS(grp[1], Env.comm[0]);
		Env.log.log_msg(Logger.INFO, "Handshake sent.");
	}

	/** Check if this group is active
	 * @return True if active, false otherwise
	 */
	@dont public boolean running ()
	{
		return !terminated;
	}

	/** Terminate the group */
	@dont public void terminate ()
	{
		terminated = true;
	}

	/** Next ID */
	@dont public String nextID ()
	{
		return myID+1+":"+members;
	}

	/** Retrieve local Logical ID
	 * @return Logical ID
	 */
	@dont public int myID ()
	{
		return myID;
	}

	/** Retrieve number of group members
	 * @return Number of members
	 */
	@dont public int members ()
	{
		return members;
	}

	/** Retrieve a string representation of this object
	 * @return String
	 */
	@dont public String toString ()
	{
		return myID+"/"+members;
	}
}
