package CTL;

import CTL.Comm.*;
import CTL.Streams.*;
import CTL.Types.*;

import java.io.*;
import java.lang.reflect.*;

/** Static methods for remote communication based on the CTL protocol */
public class Remote
{
	/** Magic string to check for correct byte order and filtering when
	 * using the pipe transport mechanism */
	public final static String magic = 
		"1F3E:A28E:2CF0:9378:AA01:0744:5D31:710A";
	public final static int magic2 = 0x1d920d4f;
	/** End of communication */
	public final static int EOC = -1;
	/** Send data directly */
	public final static int DAT = 1;
	/** Operation */
	public final static int OPER = 2;
	/** Remote method invocation */
	public final static int RMI = 3;
	/** Control message */
	public final static int CTRL = 4;
	/** Error */
	public final static int ERR = 5;
	/** Undefined */
	public final static int UNDEF = 255;
	
	/** Query (experimental) */
	protected final static int QRY = 253;
	/** List (experimental) */
	protected final static int LST = 254;

	/** If true, *Dbg calls will print out a hexdump of the streams */
	private final static boolean dump = false;
	
	/** Start tag for user defined message types */
	private static int tag = UNDEF+1;
	
	/** Convert a numerical message type to string
	 * @param mtype Message type
	 * @return String
	 */
	public static String parseMType (int mtype)
	{
		switch (mtype)
		{
			case DAT:
				return "DAT";
			case OPER:
				return "OPER";
			case RMI:
				return "RMI";
			case ERR:
				return "ERR";
			case UNDEF:
				return "UNDEF";
				
			case QRY:
				return "QRY";
			case LST:
				return "LST";
		}

		return "Other";
	}

	/** Read a DAT message
	 * @param comm Communicator to use
	 * @param type Type to read
	 * @return Read value
	 */
	public static Object readDAT (Communicator comm, String type) 
		throws IOException, IllegalAccessException, InvocationTargetException, 
		ClassNotFoundException, InstantiationException, 
		CTLException
	{
		BuffyIn in = new BuffyIn(comm.recv());
		Header head = in.i.serialRead(Header.class);
		if (head.tag()!=DAT)
		{
			String msg = "DAT expected, but "+parseMType(head.tag())+
				" received.";
			Env.log.log_msg(Logger.ERR, msg);
			throw new RuntimeException(msg);
		}
		
		if (type.equals("HEAD"))
			return head;
		Method m = RUtil.getReader(type);
		return m.invoke(in.i, (Object[])null);
	}
	
	/** Write a DAT message
	 * @param comm Communicator to use
	 * @param data Value to send
	 */
	public static void writeDAT (Communicator comm, Object data) 
		throws IOException, IllegalAccessException, InvocationTargetException, 
		ClassNotFoundException, InstantiationException, 
		CTLException
	{
		writeDAT(comm, data, comm.pid().host(), comm.pid().port());
	}

	public static void writeDAT (Communicator comm, Object data, IPaddr host,
		int port) throws IOException, IllegalAccessException, 
		InvocationTargetException, ClassNotFoundException, 
		InstantiationException, CTLException
	{
		BuffyOut out = new BuffyOut();
		Header head = new Header(0, DAT, host, port, -1);
		out.o.serialWrite(head);
		if (data != null)
			out.o.serialWrite(data);
		out.o.flush();
		comm.send(out.getBytes());
	}

	/** Write handshake
	 * @param grp Group information
	 * @param comm Communicator to use
	 */
	public static void writeHS (GroupInfo grp, Communicator comm) 
		throws IOException, IllegalAccessException, InvocationTargetException, 
		ClassNotFoundException, InstantiationException, 
		CTLException
	{
		BuffyOut out = new BuffyOut();
		Header head = new Header(28, OPER, comm);
		out.o.serialWrite(head);
		out.o.serialWrite(grp);
		out.o.flush();
		comm.send(out.getBytes());
	}

	/** Read handshake 
	 * @param comm Communicator to use
	 * @return Group information
	 */
	public static GroupInfo readHS (Communicator comm) throws IOException,
		IllegalAccessException, InvocationTargetException, 
		ClassNotFoundException, InstantiationException, 
		CTLException
	{
		BuffyIn in = new BuffyIn(comm.recv());
		Header ohead = in.i.serialRead(Header.class);
		GroupInfo ret = in.i.serialRead(GroupInfo.class);
		return ret;
	}

	/** Remote call function
	 * @param comm Communicator to use
	 * @param head Header
	 * @param objID ObjectID
	 * @param fid FunctionID
	 * @param args IStream2 which stores the arguments
	 * @param keep_args Whether or not to keep the args in the IStream2 for
	 * successive function calls
	 */
	private static void __call (Communicator comm, Header head, long objID, 
		FID fid, IStream2 args, rPointer objID2, boolean keep_args) 
		throws IOException, IllegalAccessException, InvocationTargetException,
		CTLException
	{
		BuffyOut out = new BuffyOut();
		out.o.serialWrite(head);
		out.o.writeLong(objID);
		out.o.serialWrite(fid);
		if (args!=null)
		{
			args.flush();
			args.writeToStream(out.o, keep_args);
		}
		out.o.serialWrite(objID2);
		out.o.writeInt(tag);
		out.o.flush();
		comm.send(out.getBytes(), head.size());
	}
	
	/** Call wrapper for logging, etc. */
	public static void call (Communicator comm, Header head, long objID,
		FID fid, IStream2 args, rPointer objID2) throws IOException,
		IllegalAccessException, InvocationTargetException, CTLException
	{
		if ((comm==null)||(head==null)||(fid==null)||(objID2==null))
		{
			Env.log.log_msg(Logger.ERR, "Critical Error: Null argument.");
			return;
		}
		
		int argsz = ((args!=null) ? args.streamSize() : 0);
		head.setSize(8			// objID
			+fid.size()			// fid
			+argsz				// args
			+objID2.size()		// objID2
			+4);				// tag
			
		if (!dump)
		{
			if (args!=null)
				args.flush();
			
			Env.log.log_msg(Logger.DBG, ".:. RMI call .:.");
			Env.log.log_msg(Logger.DBG, "\tHeader: "+head);
			Env.log.log_msg(Logger.DBG, "\tObjID: "+objID);
			Env.log.log_msg(Logger.DBG, "\tFID: "+fid);
			Env.log.log_msg(Logger.DBG, "\tArguments ("+argsz+"): "+args);
			Env.log.log_msg(Logger.DBG, "\trResult: "+objID2);
			Env.log.log_msg(Logger.DBG, "\tTag: "+tag);

			if (args!=null)
				args.flush();
		}
		else
		{
			// TODO: This does not work anymore
			/*__call(comm, head, objID, fid, args, objID2, true);
			Debugger.RMIdump("call", head, baos);*/
		}
			
		__call(comm, head, objID, fid, args, objID2, false);
		tag++;
	}

	/** Answer wrapper for logging, etc. */
	public static void answer (Communicator comm, Header head, long objID,
		Except ex, IStream2 args) throws IOException, IllegalAccessException,
		InvocationTargetException, CTLException
	{
		if ((comm==null)||(head==null)||(ex==null))
		{
			Env.log.log_msg(Logger.ERR, "Critical Error: Null argument.");
			return;
		}

		int argsz = ((args!=null) ? args.streamSize() : 0);
		head.setSize(8		// objID
			+ex.size()		// except
			+argsz);		// args
		
		if (!dump)
		{
			if (args!=null)
				args.flush();
		
			Env.log.log_msg(Logger.DBG, ".:. RMI answer .:.");
			Env.log.log_msg(Logger.DBG, "\tHeader: "+head);
			Env.log.log_msg(Logger.DBG, "\tObjID: "+objID);
			Env.log.log_msg(Logger.DBG, "\tException: "+ex);
			Env.log.log_msg(Logger.DBG, "\tArguments ("+argsz+"): "+args);

			if (args!=null)
				args.flush();
		}
		else
		{
			// TODO: doesnt work anymore
			/*__answer(out2, head, objID, ex, args, true);
			Debugger.RMIdump("answer", head, baos);*/
		}
			
		__answer(comm, head, objID, ex, args, false);
	}

	/** Remote answer function
	 * @param comm Communicator to use
	 * @param head Header
	 * @param objID ObjectID
	 * @param ex Exception
	 * @param args IStream2 which stores the arguments
	 * @param keep_args Whether or not to keep the arguments in the IStream2
	 * for successive calls
	 */
	private static void __answer (Communicator comm, Header head, 
		long objID, Except ex, IStream2 args, boolean keep_args) 
		throws IOException, IllegalAccessException, InvocationTargetException,
		CTLException
	{
		BuffyOut out = new BuffyOut();
		out.o.serialWrite(head);
		out.o.writeLong(objID);
		out.o.serialWrite(ex);
		if (args!=null)
		{
			args.flush();
			args.writeToStream(out.o, keep_args);
		}
		out.o.flush();
		comm.send(out.getBytes(), head.size());
	}
}
