package CTL.Streams;

import CTL.*;
import CTL.Serialize.*;
import CTL.Types.*;
import ReflWrap.*;

import java.io.*;
import java.util.*;
import java.lang.reflect.*;

/** "Intelligent" Stream: This special stream is handled just like an
 * ObjectOutputStream, however, it saves information about the data which was
 * written to it. Therefore, it is an easy interface to pass around data
 * from different sources to one ObjectOutputStream and it is also capable of
 * writing a "signature" string to the stream which can be interpreted by the
 * readFromStream() call. This makes it possible to pass around larger amounts
 * of arbitrary data via Object*putStreams w/o recreating the structure of
 * the stream in code on both ends. It is also possible to query the size of
 * the whole stream. Furthermore, the stream can write the data to other
 * streams while keeping the data inside, which makes it ideal for debugging
 * stream related applications.
*/
public class IStream
{
	/** LinkedList of types of the stored data */
	private LinkedList<Class> list;
	/** Input stream for the internal storage */
	private SerialIn in;
	/** Output stream for the internal storage */
	private SerialOut out;
	/** Size of the stream */
	private int size;
	
	/** Create a new IStream */
	public IStream () throws IOException, CTLException
	{
		reset();
	}

	/** Reset the internal storage (via Piped*Streams) */
	public void reset () throws IOException, CTLException
	{
		PipedOutputStream pout = new PipedOutputStream();
		PipedInputDream pin = new PipedInputDream(pout);
		
		out = new OOStream(pout);
		in = new OIStream(pin);
		list = new LinkedList<Class>();
		size = 0;
	}

	/** Flush the stream */
	public void flush () throws IOException
	{
		out.flush();
	}

	/** Close the stream */
	public void close () throws IOException
	{
		out.close();
	}

	/** Write data to the stream
	 * @param val Value
	 */
	public <T extends Object> void write (T val) throws IOException,
		IllegalAccessException, InvocationTargetException, CTLException
	{
		if (val==null) return;
		
		list.add(val.getClass());
		
		Object[] args = new Object[1];
		args[0] = val;
		RUtil.serialwr.invoke(out, args);
		
		size+=RUtil.objSize(val);
	}
		
	/** Write data to an OOStream
	 * @param outp Stream to write to
	 * @param stuff Data to write
	 */
	public static <T extends Object> void writeType (SerialOut outp, T stuff)
		throws IllegalAccessException, InvocationTargetException
	{
		Object[] args = new Object[1];
		args[0] = stuff;
		Method m = RUtil.serialwr;
		m.invoke(outp, args);
	}

	/** Read data from an OIStream
	 * @param input Stream to read from
	 * @param typ Type to read
	 * @return Data read
	 */
	public static Object readType (SerialIn input, Class typ) 
		throws IllegalAccessException, InvocationTargetException
	{
		String type = typ.getName();
		if (type.substring(0, 1).equals("["))
			type = Refl.type(type);
		Method m = RUtil.getReader(type);
		Object[] args = null;

		// Array type
		if (type.charAt(type.length()-1)==']')
		{
			args = new Object[1];
			args[0] = new ClassInfo(typ).arrayType();
		}	
		
		// Composite type
		if (m.getName().equals("readObject"))
		{
			args = new Object[1];
			try
			{
				args[0] = Class.forName(type);
			} catch (Exception e) {}
			m = RUtil.serialrd;
		}
		
		Env.log.log_msg(Logger.DBG2, "IStream.readType using "+m);
		return m.invoke(input, args);
	}

	/** Read data from this stream
	 * @param type Type to read
	 * @return Data read 
	 */
	private Object readType (Class type)
		throws IllegalAccessException, InvocationTargetException
	{
		return readType(in, type);
	}

	/** Retrieve a string representation of this object
	 * @return String
	 */
	public String toString ()
	{
		StringBuffer ret = new StringBuffer();
		
		try
		{
			// XXX: Reference map is static.
			Reference.cleanup();

			LinkedList<Object> content = new LinkedList<Object>();
			LinkedList<Class> types = new LinkedList<Class>(list);
			
			for (Class c : list)
			{
				Object foo = readType(c);
				content.add(foo);
				if ((foo!=null) && foo.getClass().isArray())
					for (int i=0;i<java.lang.reflect.Array.getLength(foo);i++)
						ret.append(java.lang.reflect.Array.get(foo, i)+" ");
				else
					ret.append(foo);
				ret.append("\n");
			}					
			
			if (ret.length()>0)
				ret.deleteCharAt(ret.length()-1);
			reset();

			for (Object o : content)
				write(o);

			flush();
			list = types;
		}
		catch (Exception e)
		{
			RUtil.except(e);
		}
		
		return ret.toString();
	}

	/** Write the contents of this stream to another one 
	 * @param ostr Target stream
	 * @param keep Whether to keep the data in this stream
	 */
	public void writeToStream (SerialOut ostr, boolean keep)
	{
		try
		{
			// XXX: Reference map is static.
			Reference.cleanup();
			
			LinkedList<Object> content = new LinkedList<Object>();
			LinkedList<Class> types = new LinkedList<Class>(list);

			for (Class c : list)
			{
				Object arg = readType(c);
				writeType(ostr, arg);
				if (keep)
					content.add(arg);
			}

			reset();
			
			if (keep)
			{
				for (Object o : content)
					write(o);
			}
			
			flush();
			list = types;
		}
		catch (Exception e)
		{
			RUtil.except(e);
		}
	}

	/** Write the data of this stream to another without keeping it
	 * @param ostr Target stream
	 */
	public void writeToStream (SerialOut ostr)
	{
		writeToStream(ostr, false);
	}

	/** Write the data of this stream to another IStream while keeping it
	 * @param ostr Target stream
	 */
	public void writeToIStream (SerialOut ostr)
	{
		writeToIStream(ostr, true);
	}

	/** Write the data of this stream to another IStream
	 * @param ostr Target stream
	 * @param keep True if the data should be kept, false otherwise.
	 */
	public void writeToIStream (SerialOut ostr, boolean keep)
	{
		try
		{
			StringBuffer buf = new StringBuffer();
			for (Class c : list)
				buf.append(c.getName()+"$$");
			buf.delete(buf.length()-2, buf.length());
			ostr.writeString(buf.toString());
			writeToStream(ostr, keep);
		}
		catch (Exception e)
		{
			RUtil.except(e);
		}
	}

	/** Read data from another IStream
	 * @param istr Source stream
	 */
	public void readFromStream (SerialIn istr)
	{
		try
		{
			LinkedList<Class> types = new LinkedList<Class>();
			StringTokenizer tok = new StringTokenizer(istr.readString(), "$$");
			while (tok.hasMoreTokens())
			{
				String type = tok.nextToken();
				types.add(Class.forName(type));
				Class moo = Class.forName(type);
				write(readType(istr, moo));
			}
			list = types;
		}
		catch (Exception e)
		{
			RUtil.except(e);
		}
	}
		
	/** Retrieve the size of the data this stream holds
	 * @return Size
	 */
	public int streamSize ()
	{
		return size;
	}
}
