package CTL.Types;

import CTL.*;
import CTL.Serialize.*;
import CTL.Streams.*;

import com.jcraft.jsch.*;
import java.io.*;
import java.lang.reflect.*;
import java.util.regex.*;
import java.util.LinkedList;

/** Location of a remote CTL service */
public class Location implements Writable
{
	public final static int FILE 	= 	0;
	public final static int SERIAL 	= 	1;
	public final static int MPI		=	2;
	public final static int LAM		=	3;
	public final static int PVM		=	4;
	public final static int THREAD	=	5;
	public final static int TCP		=	6;
	public final static int LIB		=	7;
	public final static int DMN		=	8;
	public final static int SSH		=	9;
	public final static int PIPE	=	10;
	public final static int UNDEF	=	63;
	public final static int HTTP	=	70;
	
	/** Command which starts the service */
	protected String cmd;
	/** Path to the executable */
	protected String path;
	/** Remote host */
	protected String host;
	/** Port (22 is default for SSH) */
	protected int port = -1;
	/** Username */
	public String user;
	/** Password (will be prompted if not specified */
	private String pass;
	/** Linkage */
	protected int linkage;

	/** Retrieve linkage
	 * @return Linkage
	 */
	public int linkage ()
	{
		return linkage;
	}

	/** Convert linkage to string
	 * @param linkage Linkage
	 * @return String
	 */
	private static String linkage2str (int linkage)
	{
		switch (linkage)
		{
			case FILE:
				return "file";
			case SERIAL:
				return "serial";
			case MPI:
				return "mpi";
			case LAM:
				return "lam";
			case PVM:
				return "pvm";
			case THREAD:
				return "thread";
			case TCP:
				return "tcp";
			case LIB:
				return "lib";
			case DMN:
				return "dmn";
			case SSH:
				return "ssh";
			case PIPE:
				return "pipe";
			case HTTP:
				return "http";
		}
		return "undef";
	}

	/** Retrieve command
	 * @return Command
	 */
	public String cmd ()
	{
		return cmd;
	}

	/** Retrieve path
	 * @return Path
	 */
	public String path ()
	{
		return path;
	}

	/** Check if a password was given
	 * @return True if yes, false otherwise.
	 */
	public boolean hasPassword ()
	{
		return pass!=null;
	}

	/** Set the password for an SSHv2 session
	 * @param sess Session
	 */
	public void setPassword (Session sess)
	{
		if (hasPassword())
			sess.setPassword(pass);
	}	

	/** Read from stream
	 * @param in Stream to read from
	 */
	public void read (SerialIn in) throws IOException, ClassNotFoundException, 
		InstantiationException, IllegalAccessException, 
		InvocationTargetException
	{
		cmd = in.readString();
		path = in.readString();
		host = in.readString();
		user = in.readString();
	}
	
	/** Write to stream
	 * @param out Stream to write to
	 */
	public void write (SerialOut out) throws IOException, IllegalAccessException,
		InvocationTargetException
	{
		out.writeString(cmd);
		out.writeString(path);
		out.writeString(host);
		out.writeString(user);
	}

	/** Location constructor */
	public Location (String cmd, String path, String host, int port, 
		String user, String pass, int linkage)
	{
		this.cmd = cmd;
		this.path = path;
		this.host = host;
		this.port = port;
		this.user = user;
		this.pass = pass;
		this.linkage = linkage;
	}

	/** Location constructor with default port */
	public Location (String cmd, String path, String host, String user,
		String pass, int linkage)
	{
		this(cmd, path, host, 22, user, pass, linkage);
	}

	/** Location constructor with default linkage and port */
	public Location (String cmd, String path, String host, String user,
		String pass)
	{
		this(cmd, path, host, 22, user, pass, TCP);
	}

	/** Location constructor with default port and password prompt */
	public Location (String cmd, String path, String host, String user,
		int linkage)
	{
		this(cmd, path, host, 22, user, Password.prompt(), linkage);
	}

	/** Location constructor for a service which is already running */
	public Location (String host, int port, int linkage)
	{
		this(null, null, host, port, null, null, linkage);
	}

	/** Location constructor from a string
	 * @param loc String
	 */
	public Location (String loc)
	{
		this(null, null, null, 0, null, null, TCP);
		if (loc.equals(""))
			return;
		
		Pattern p = Pattern.compile(" ");
		String[] foo = p.split(loc);

		if ((foo.length > 1) && foo[1].equals("dmn"))
		{
			p = Pattern.compile(":");
			String[] bar = p.split(foo[0]);
			host = bar[0];
			port = Integer.parseInt(bar[1]);
			linkage = DMN;
			return;
		}
		
		if ((foo.length > 1) && (foo[1].equals("tcp") || foo[1].equals("pipe")))
		{
			if (foo[1].equals("tcp"))
				linkage = TCP;
			else
				linkage = PIPE;
			p = Pattern.compile("@");
			Matcher m = p.matcher(foo[0]);
			if (m.find())
			{
				String[] moo = p.split(foo[0]);
				foo[0] = moo[1];
				p = Pattern.compile(":");
				moo = p.split(moo[0]);
				user = moo[0];
				if (moo.length > 1)
					pass = moo[1];
			}
			p = Pattern.compile(":");
			String[] bar = p.split(foo[0]);
			host = bar[0];
			port = 22;
			p = Pattern.compile("/");
			foo = p.split(bar[1]);
			path = "";
			// Absolute path assumed.
			for (int i=1;i<foo.length-1;i++)
				path += "/"+foo[i];
			cmd = foo[foo.length-1];
			return;
		}

		throw new RuntimeException("Could not parse location string '"+
				loc+"'.");
	}

	/** Retrieve a string representation of this object
	 * @return String
	 */
	public String toString ()
	{
		String ret = linkage2str(linkage) + ": ";
		if (user != null)
		{
			ret += user;
			// Not printing passwords for security reasons.
			/*if (pass != null)
				ret += ":"+pass;*/
			ret += "@";
		}
		ret += host;
		if (port > 0)
			ret += ":"+port;
		ret += " ";
		if (path != null)
			ret += path + " ";
		if (cmd != null)
			ret += cmd;
		return ret;
	}

	/** Retrieve the hostname
	 * @return Hostname
	 */
	public String host ()
	{
		return host;
	}

	/** Retrieve the port
	 * @return Port
	 */
	public int port ()
	{
		return port;
	}

	/** Read a bunch of locations from a file 
	 * @param fname Filename
	 * @return Array of locations
	 */
	public static LinkedList<Location> parseFile (String fname)
	{
		try
		{
			return parseFile(new File(fname));
		} catch (IOException e) {}
		return null;
	}

	/** Read a bunch of locations from a file
	 * @param file File
	 * @return Array of locations
	 */
	public static LinkedList<Location> parseFile (File file) throws IOException
	{
		LinkedList<Location> list = new LinkedList<Location>();
		BufferedReader buf = new 
			BufferedReader(new FileReader(file));
		while (buf.ready())
		{
			String line = buf.readLine();
			if (line.charAt(0) == '#')
				continue;
			list.add(new Location(line));
		}
		buf.close();
		return list;
	}
}
