package CodeGen;

import CTL.*;
import CTL.Annotate.*;
import ReflWrap.*;

import java.io.*;
import java.lang.annotation.*;
import java.lang.reflect.*;
import java.util.*;

/** Code generator for Java remote interfaces using Reflection */
public class GenRI extends Gen
{
	private ClassInfo cinfo;
	private String clname, pkg, prfx;
	
	/** Constructor for a class from the javaSys package.
	 * @param klass Class name
	 */
	public GenRI (String klass) throws ClassNotFoundException
	{
		this("javaSys", klass);
	}
	
	/** Constructor for a class from a custom package.
	 * @param pkg Package name
	 * @param klass Class name
	 */
	public GenRI (String pkg, String klass) throws ClassNotFoundException
	{
		this.pkg = pkg;
		cinfo = new ClassInfo(Class.forName(klass));

		if (cinfo._package().equals("_default"))
			this.pkg = "_default";
		
		if (this.pkg.equals("_default"))
			prfx = "";
		else
			prfx = this.pkg+"::";
		
		String name = cinfo.name();
		if (ByteCode.matches(name, ".*I$"))
			clname = name.substring(0, name.length()-1);
		else
			clname = name+"RI";
	}
	
	/** Generate a remote interface */
	protected void genRI () throws IOException, NoSuchMethodException
	{
		FileOutputStream fos = null;
		
		try
		{
			// Check if the file was modified recently
			File impl = new File(cinfo.filename());
			File ri = new File("src/"+pkg+"/"+clname+".java");
		
			if (impl.lastModified()<ri.lastModified())
			{
				File gen = new File("CodeGen/GenRI.java");
				if (gen.lastModified()<ri.lastModified())
					return;
			}
			
			ri.delete();
			fos = new FileOutputStream(ri);
			gen(fos);
			prettyPrint(ri);
		}
		finally
		{
			if (fos!=null)
				fos.close();
		}
	}
	
	/** Generate a remote interface
	 * @param out RI will be written to this
	 */
	protected void gen (OutputStream out) throws IOException, 
		NoSuchMethodException
	{
		OutputStreamWriter ostr = new OutputStreamWriter(out);
		
		writeTo(ostr, genHead());
		writeTo(ostr, genConstructors());
		writeTo(ostr, genWrappers());
		writeTo(ostr, genAccept());
		writeTo(ostr, "}\n");

		ostr.flush();
	}

	/** Generate a *Value() call for a certain type
	 * @param retVal String representation of a type
	 * @return Array of which the first String is the *Value() function
	 * name
	 */
	protected static String[] genValue (String retVal) 
	{
		String[] ret = new String[2];
		Class foo = Refl.getClass(Refl.convert(retVal));

		if (retVal.equals("Array"))
			retVal = "Object";
		
		ret[1] = "";
		
		if ((retVal!=null)&&(foo==null))
		{	
			ret[1] = "("+retVal+")";
			retVal = "Object";
		}
	
		ret[0] = retVal+"Value";

		return ret;
	}

	/** Generate the header of the class
	 * @return Header
	 */
	protected String genHead () throws IOException
	{
		StringBuffer ret = new StringBuffer();

		ret.append("package "+pkg+";\n\n");
		ret.append("import CTL.*;\n");
		ret.append("import CTL.CCompat.*;\n");
		ret.append("import CTL.Comm.*;\n");
		ret.append("import CTL.Streams.*;\n");
		ret.append("import CTL.Types.*;\n");
		ret.append("import ReflWrap.*;\n\n");
		ret.append("import java.io.*;\n");
		ret.append("import java.net.*;\n");
		ret.append("import java.lang.reflect.*;\n");
		if (cinfo._package().equals(""))
		{
			System.err.println("Implementation belongs to no package.");
			System.exit(1);
		}
		else
			ret.append("import "+cinfo._package()+".*;\n");
		getImports(ret, cinfo.fqcn());
		ret.append("\n/* This class was autogenerated by the CTL4j code"+
			" generator from\n *\t\t"+cinfo.fqcn()+".\n * Written by Boris"+
			" Buegling, licensed under the GNU General Public license. */\n");
		ret.append("public class "+clname+" extends RI\n");
		ret.append("{\n");

		return ret.toString();
	}

	/** Generate constructor wrappers
	 * @return Code
	 */
	protected String genConstructors () throws IOException
	{
		StringBuffer ret = new StringBuffer();

		for (int i=0;i<cinfo.constructors().length;i++)
		{
			Constructor cnst = cinfo.constructors()[i];
			if (cnst.isAnnotationPresent(cdont.class))
				continue;
			ConstructInfo cInfo = new ConstructInfo(cnst);
			boolean _const = cinfo.isAnot(const_.class);
			String params = cInfo.paramsToString();
			int len = cInfo.params().length;
			
			ret.append("public "+clname+" ("+params+"\n");
			ret.append("{\nthis(proc()"+((len==0) ? "" : ", "));
			for (int j=0;j<len;j++)
				ret.append("arg"+j+((j==len-1) ? "" : ", "));
			ret.append(");\n}\n");
			
			ret.append("public "+clname+" (CTL.Process proc"+
				((len==0) ? "" : ", ")+params+"\n");
			ret.append("{\ntry\n{\n");
			ret.append("FID fid = new FID((short)"+i+", \""+prfx+
					clname+":"+i+"C\");\n");
			ret.append("Constructor c = rResult.class.getConstructors()[0];\n");
			ret.append("Object[] objs = new Object[1];\n");
			ret.append("objs[0] = (Object)new TypeTree(rPointer.class);\n");
			ret.append("long objID = Env.map.createObj(c, objs);\n");
			ret.append("IStream2 args = new IStream2();\n");
			
			for (int j=0;j<len;j++)
			{
				ret.append("args.write(arg"+j+");\n");
			}
			
			ret.append("Communicator comm = Env.comm[proc.ID()];\n");
			ret.append("Header head = new Header(0, Remote.RMI, comm);\n");
			ret.append("rPointer rp = new rPointer(objID, comm);\n");
			ret.append("Remote.call(comm, head, 0, fid, args, rp);\n");
			ret.append("rResult foo = (rResult)Env.map.getObj((int)objID);\n");
			ret.append("if (foo==null)\n{\n");
			ret.append("Env.log.log_msg(Logger.ERR, \"No data was "+
				"received.\");\n");
			ret.append("throw new RuntimeException(\"No data was received."+
				"\");\n}\n");
			ret.append("while (!foo.received(foo.results()-1));\n");
			ret.append("setObjID(((rPointer)foo.ObjectValue("+
				"foo.results()-1)).objID());\n");
			ret.append("Env.log.log_msg(Logger.ERR, \"Object created with "+
				"ObjectID \"+objID()+\".\");\n");
			ret.append("}\ncatch (Exception e)\n{\n");
			ret.append("Env.log.log_msg(Logger.ERR, \"Constructor "+
				"wrapper:\");\n");
			ret.append("RUtil.except(e);\n}\n}\n");
		}

		return ret.toString();
	}

	// TODO: Cleanup generated code for void methods
	/** Generate function wrappers
	 * @return Code
	 */
	protected String genWrappers () throws IOException, NoSuchMethodException
	{
		int me=1, s=1;
		StringBuffer ret = new StringBuffer();

		for (int i=0;i<cinfo.methods().length;i++)
		{
			Method moo = cinfo.methods()[i];
			// If there are unassigned FIDs, skip them.
			if (moo == null)
				continue;
			MethodInfo minfo = new MethodInfo(moo);
			boolean _const = cinfo.isAnot(const_.class );
			
			if (moo.isAnnotationPresent(dont.class ))
			{
				if (minfo.isStatic())
					me++;
				else
					s++;
				
				continue;
			}
				
			int len = minfo.params().length;
			if (minfo._class()!=cinfo._class())
				continue;

			/*
			 * Two wrappers are created:
			 * 	1. Simple wrapper
			 * 	2. rResult wrapper
			 */
			for (int rres=0;rres<2;rres++)
			{
				boolean rr = rres==1;
				boolean isvoid = minfo.returns().name().equals("void");

				if (isvoid && rr)
					continue;
				
				// Method declaration	
				if (rr)
					ret.append(minfo.modifiersToString()+
						(isvoid ? "void " : "rResult ")+
						 minfo.name()+"_rr ("+minfo.paramsToString());
				else
					ret.append(minfo);
				
				/* 
				 * Static methods can be called with an optional CTL.Process 
				 * argument
				 */
				if (minfo.isStatic())
				{
					ret.append("\n{\n"+
						(isvoid ? "" : "return ")+
						minfo.name()+(rr ? "_rr" : "")+"(proc()"+
						((len==0) ? "" : ", "));
					for (int j=0;j<len;j++)
						ret.append("arg"+j+((j==len-1) ? "" : ", "));
					ret.append(");\n}\n"+minfo.modifiersToString()+
						(rr ? "rResult" : minfo.returns().name())+
						" "+minfo.name()+(rr ? "_rr" : "")+
						" (CTL.Process proc"+((len==0) ? "" : ", ")+
						minfo.paramsToString()+"\n");
				}
			
				/*
				 * Type of the return value is either the actual return type
				 * or a rResult.
				 */
				ret.append("\n{\n");
				if (!isvoid)
					if (rr)
						ret.append("rResult res = null;\n");
					else
						ret.append(minfo.returnsT()+" res = "+
							Refl.defaultVal(minfo.returns().name())+";\n");
				
				if (minfo.isStatic())
					ret.append("try\n{\n"+"if (proc==null)\n{\n"+
						"Env.log.log_msg(Logger.ERR, "+
							"\"No remote process locator defined.\");\n"+
						"throw new RuntimeException(\"No remote process "+
						"locator defined.\");\n}\n");
				else
					ret.append("CTL.Process proc = proc();\n"+
						"try\n{\n"+"if (objID()==-1)\n"+
						"throw new Exception(\"ObjectID not set!\");\n");
			
				ret.append("Constructor c = "+
						"rResult.class.getConstructors()[1];\n");
				
				// Code for the rResult construction
				int nret=0;
				for (int k=0;k<minfo.getParams().length;k++)
					if (!Refl.hasAnot("@CTL.Annotate.const_()", 
								minfo.getParamAnot(k)) && !_const)
						nret++;
				ret.append("TypeTree[] argz = new TypeTree["+(nret+1)+"];\n");
				int p=0;
				for (int k=0;k<minfo.getParams().length;k++)
					if (!Refl.hasAnot("@CTL.Annotate.const_()", 
							minfo.getParamAnot(k)) && !_const)
					{
						String arg = Refl.class_(minfo.getParams()[k]);
						if (ByteCode.matches(minfo.params()[k], ".*<.*"))
							arg = "\""+minfo.params()[k]+"\"";
						ret.append("argz["+(p++)+"] = new TypeTree("+
							arg+");\n");
					}

				String arg = Refl.class_(minfo.returns()._class());
				if (ByteCode.matches(minfo.returnsT(), ".*<.*"))
					arg = "\""+minfo.returnsT()+"\"";
				ret.append("argz["+nret+"] = new TypeTree("+arg+");\n"+
					"Object[] objs = new Object[1];\n"+
					"objs[0] = (Object)argz;\n"+
					"long objID = Env.map.createObj(c, objs);\n"+
					"IStream2 args = new IStream2();\n");
			
				for (int j=0;j<len;j++)
					ret.append("args.write(arg"+j+");\n");

				ret.append("Communicator comm = Env.comm[proc.ID()];\n");
				ret.append("FID fid = new FID((short)");
				
				if (minfo.isStatic())
				{
					ret.append(me+", \""+
						prfx+clname+":"+me+"S\");\n"+
						"Header head = new Header(0, Remote.RMI, comm);\n"+
						"rPointer rp = new rPointer(objID, comm);\n"+
						"Remote.call(comm, head, 0, fid, args, rp);\n\n");
					me+=rres;
				}
				else
				{
					ret.append(s+", \""+"\");\n"+
						"Header head = new Header(0, Remote.RMI, comm);\n"+
						"rPointer rp = new rPointer(objID, comm);\n"+
						"Remote.call(comm, head, objID(), fid, args, rp);\n\n");
					s+=rres;
				}

				if (!isvoid)
				if (rr)
					ret.append("res = (rResult)Env.map.getObj((int)objID);\n");
				else
				{
					ret.append("rResult foo = (rResult)Env.map.getObj((int)"+
						"objID);\n"+
						"while (!foo.received(foo.results()-1));\n");

					String[] val = genValue(Refl.rtype(minfo.returns()._class()));
					Class[] args = new Class[1];
					args[0] = Integer.TYPE;
					Method m = rResult.class.getDeclaredMethod(
						Refl.lastToken(val[0], "."), args);
					ret.append("res = ");
					if (minfo.returns().isArray())
						ret.append("Refl."+minfo.returns().arrayType().getName()+
							"Array(");
					ret.append(val[1]+"foo."+m.getName()+"(foo.results()-1)");
					if (minfo.returns().isArray())
						ret.append(")");
					ret.append(";\n");
					ret.append("Env.log.log_msg(Logger.DBG2, \"rResult: \""+
						"+(foo.results()-1)+\" used.\");\n");

					int bz=0;
					for (int k=0;k<minfo.getParams().length;k++)
						if (!Refl.hasAnot("@CTL.Annotate.const_()", 
									minfo.getParamAnot(k)) && !_const)
						{
							String[] val2 = genValue(Refl.rtype(
								minfo.getParams()[k]));
							Class[] args2 = new Class[1];
							args2[0] = Integer.TYPE;
							Method muh = rResult.class.getDeclaredMethod(
								val2[0], args2);
							ret.append("Env.log.log_msg(Logger.DBG, "+
								"\"non-const arg "+k+" is \"+foo."+
								muh.getName()+"("+k+"));\n");
						}
				}
				
				ret.append("}\ncatch (Exception e)\n{\n"+
					"Env.log.log_msg(Logger.ERR, \""+
					(rr ? "rResult w" : "W")+"rapper:\");\n"+
					"RUtil.except(e);\n}\n");
		
				if (!isvoid)
					ret.append("return res;\n");
				
				ret.append("}\n");
			}
		}

		return ret.toString();
	}

	/** Generate the accept() function
	 * @return Code
	 */
	protected String genAccept () throws IOException
	{
		StringBuffer ret = new StringBuffer();
		StringBuffer ret2 = new StringBuffer();
		int j, st=1, me=1;
		boolean _const = cinfo.isAnot(const_.class);

		ret.append("@SuppressWarnings(\"unchecked\")\n");
		ret.append("public static void accept (OIStream in, Header ohead, long objID, FID fid, int id) throws IOException, ClassNotFoundException, InstantiationException, IllegalAccessException, NoSuchMethodException, InvocationTargetException, CTLException\n");
		ret.append("{\nClass[] args2 = null;\n");
		ret.append("Header head = null;\n");
		ret.append("long ID = 0;\n");
		ret.append("int tag = 0;\n");
		ret.append("IStream2 args = null;\n");
		ret.append("rPointer rp = null;\n");
		ret.append("OOStream out = null;\n");
		ret.append("IPaddr host = Env.comm[0].pid().host();\n");
		ret.append("int port = Env.comm[0].pid().port();\n");
		ret.append("Env.log.log_msg(Logger.DBG2, ohead+\" \"+objID);\n");
		ret.append("if (objID==0)\n{\n");
		ret.append("char c = fid.name().charAt(fid.name().length()-1);\n");
		ret.append("Env.log.log_msg(Logger.DBG2, new Integer(c).toString());\n");
		ret.append("Env.log.log_msg(Logger.DBG2, fid.toString());\n");	
		ret.append("switch (c)\n{\n");
		ret.append("case 'C':\n");
		ret.append("switch (fid.ID())\n{\n");

		// Constructors
		for (int i=0;i<cinfo.constructors().length;i++)
		{
			Constructor cnst = cinfo.constructors()[i];
			if (cnst.isAnnotationPresent(cdont.class))
				continue;
			ConstructInfo cinfo2 = new ConstructInfo(cnst);
			
			ret.append("case "+i+":\n");
			if (cinfo2.params().length<1)
				ret.append("Object[] foo"+i+" = null;\n");
			else
			{
				ret.append("Object[] foo"+i+" = new Object["+
					cinfo2.params().length+"];\n");
				for (j=0;j<cinfo2.params().length;j++)
					genArgList(ret, cinfo2.params()[j], i, j, true, 
						cinfo2.getParamAnot(j));	

				// TODO: Check genArgList() for constructors
				/*{
					String p = Refl.convert(cinfo2.params()[j]);
					String reader = RUtil.getReader(p).getName();
					if (p.equals("Int"))
						p = "Integer";
					ret.append("foo"+i+"["+j+"] = (Object)");
					ret.append("new "+p+"(in."+reader+"());\n");
				}*/
			}
			ret.append("rp = in.serialRead(rPointer.class);\n");
			ret.append("Env.log.log_msg(Logger.DBG, \"\"+rp);\n");
			ret.append("tag = in.readInt();\n");
			if (cinfo2.params().length<1)
			{
				ret.append("args2 = null;\n");
			}
			else
			{
				ret.append("args2 = new Class["+
					cinfo2.params().length+"];\n");
				for (j=0;j<cinfo2.params().length;j++)
				{
					String s = Refl.convert(cinfo2.params()[j]);
					ret.append("args2["+j+"] = ");
					if (Refl.getClass(s)==null)
						ret.append(s+".class");
					else
						ret.append("Refl.getClass(\""+s+"\")");
					ret.append(";\n");
				}
			}
			ret.append("Constructor c"+i+" = "+cinfo.fqcn()+".class.\n");
			ret.append("getDeclaredConstructor(args2);\n");
			ret.append("ID = Env.map.createObj(c"+i+", foo"+i+");\n");
			ret.append("args = new IStream2();\n");
			ret.append("args.write(new rPointer(ID, Env.comm[id]));\n");
			ret.append("head = new Header(0, tag, Env.comm[id]);\n");
			ret.append("Remote.answer(Env.comm[id], head, "+
				"rp.objID(), new Except(),\n");
			ret.append("args);\n");
			ret.append("break;\n");
		}
		
		ret.append("}\nbreak;\ncase 'S':\n");
		ret.append("switch (fid.ID())\n{\n");
		
		// Methods
		for (int i=0;i<cinfo.methods().length;i++)
		{
			Method moo = cinfo.methods()[i];
			// If there are unassigned FIDs, skip them.
			if (moo == null)
				continue;
			MethodInfo minfo = new MethodInfo(moo);
			
			if (moo.isAnnotationPresent(dont.class))
			{
				if (minfo.isStatic())
					st++;
				else
					me++;
				
				continue;
			}
			
			if (minfo._class()!=cinfo._class())
				continue;
				
			StringBuffer loc;
			if (minfo.isStatic())
			{
				loc = ret;
				loc.append("case "+st+":\n");
				st++;
			}
			else
			{
				loc = ret2;
				loc.append("case "+me+":\n");
				me++;
			}
			
			for (j=0;j<minfo.params().length;j++)
				genArgList(loc, minfo.params()[j], i, j, false, 
					minfo.getParamAnot(j));

			if (!moo.isAnnotationPresent(builtin.class))
			{
				loc.append("rp = in.serialRead(rPointer.class);\n");
				//loc.append("Env.log.log_msg(Logger.DBG, \"\"+rp);\n");
			}
			loc.append("tag = in.readInt();\n");
			
			if (minfo.isStatic()==true)
			{
				loc.append(((minfo.returnsT().equals("void")) ? ""
					: minfo.returnsT()+" res"+i+" = ")+cinfo.fqcn()+"."+
					minfo.name()+"(");
			}
			else
			{
				loc.append(cinfo.fqcn()+" foo"+i+" = ("+cinfo.fqcn()+
					")Env.map.getObj((int)objID);\n");
				loc.append(((minfo.returnsT().equals("void")) ? ""
					: minfo.returnsT()+" res"+i+" = ")+"foo"+
					i+"."+minfo.name()+"(");
			}
				
			for (j=0;j<minfo.params().length;j++)
				loc.append("arg"+j+i+((j==minfo.params().length-1) ? ");\n" 
					: ", "));
			if (j==0)
				loc.append(");\n");
				
			loc.append("Env.log.log_msg(Logger.DBG, \"Method "+minfo.name());
			loc.append(" was called with arguments\"+");
			for (j=0;j<minfo.params().length;j++)
				loc.append("\" arg"+j+" = "+"\"+arg"+j+i+"+\",\"+");
			if (!minfo.returns().name().equals("void"))
				loc.append("\" and result \"+res"+i+"+\".\");\n");
			else
				loc.append("\".\");\n");
			
			if (!minfo.returns().name().equals("void"))
			{
				loc.append("args = new IStream2();\n");

				if (!_const)
				{
					for (int k=0;k<minfo.getParams().length;k++)
						if (!Refl.hasAnot("@CTL.Annotate.const_()", 
								minfo.getParamAnot(k)))
							loc.append("args.write(arg"+k+i+");\n");
				}

				loc.append("args.write(res"+i+");\n");
				loc.append("head = new Header(0, tag, Env.comm[id]);\n");
				loc.append("Remote.answer(Env.comm[id], head, rp.objID(),"+
					" new Except(), args);\n");
			}
			
			loc.append("break;\n");
		}
			
		ret.append("}\nbreak;\n}\n");
		ret.append("}\nelse\n{\n");
		ret.append("Env.log.log_msg(Logger.DBG2, fid.toString());\n");
		ret.append("switch (fid.ID())\n{\n");
		ret.append(__putRefCount());
		ret.append(ret2.toString());
		ret.append("}\n}\n}\n");

		return ret.toString();
	}

	/** Handles one function parameter and takes care of possible array
	 * types.
	 * @param buf StringBuffer to write to
	 * @param p Parameter type
	 * @param i Counter
	 * @param j Counter
	 * @param is_constr Whether or not the caller is a constructor
	 * @param anot Annotations belonging to the caller
	 */
	private static void genArgList (StringBuffer buf, String p, int i, int j, 
		boolean is_constr, Annotation[] anot)
	{
		String reader = RUtil.getReader(p).getName();
		boolean is_arr = reader.equals("readArray");
		boolean foobar = p.endsWith("[]");

		if (!Refl.hasAnot("@CTL.Any()", anot) &&
			reader.equals("readObject"))
			reader="serialRead";
		
		if (is_constr)
			buf.append("foo"+i+"["+j+"] = ");
		else
			buf.append(((is_arr) ? "Object[]" : p)+" arg"+((is_arr) ? "T" : "")+
			j+i+" = "+((!Refl.convert(p).equals(RUtil.reader2str(reader)))
			? "("+p+")" : ""));

		if (foobar)
		{
			String foo = p.substring(0, p.length()-2);
			buf.append("Refl."+foo+"Array(");
		}
		
		buf.append("in."+reader+"(");
		buf.append((is_arr) ? Refl.arrayType2(p) : "");
		// TODO: Temporary hack, does not work for parametrized types
		buf.append((is_arr) ? ", null" : "");
			
		if (reader.equals("serialRead"))
		{
			String type = Refl.firstToken(p, "<");
			if (!type.equals(p))
				buf.append("new TypeTree(\""+p+"\")");
			else
				buf.append(p+".class");
		}
		
		if (foobar)
			buf.append(")");
		buf.append(");\n");
		
		// TODO: This will not work for constructors
		if (is_constr) return;
		if (is_arr)
		{
			buf.append(p+" arg"+j+i+" = new "+p.substring(0, p.length()-1)+
				"argT"+j+i+".length];\n");
			buf.append("for (int i=0;i<argT"+j+i+".length;i++)\n");
			buf.append("arg"+j+i+"[i] = "+RUtil.converter(p.substring(0,
				p.length()-2), "argT"+j+i+"[i]")+";\n");
		}
	}

	/** Generate a stub for putRefCount()
	 * @return Code
	 */
	private static String __putRefCount ()
	{
		StringBuffer buf = new StringBuffer();
		buf.append("case 0:\n");
		buf.append("int arg0 = in.readInt();\n");
		buf.append("Env.log.log_msg(Logger.ERR, \"STUB: RI.putRefCount()"+
			" (Value: \"+arg0+\")\");\n");
		buf.append("break;\n");
		return buf.toString();
	}
}
