package ReflWrap;

import java.util.*;
import java.util.regex.*;
import org.apache.bcel.*;
import org.apache.bcel.classfile.*;
import org.apache.bcel.generic.*;

/** Wrapper around the BCEL
 * @link http://jakarta.apache.org/bcel/
 */
public class ByteCode
{
	/** Sed-like search-and-replace for Strings
	 * @param input String to work on
	 * @param pat Pattern to look for
	 * @param repl Replacement
	 * @return Modified string
	 */
	public static String sed (String input, String pat, String repl)
	{
		Pattern p = Pattern.compile(pat);
		Matcher m = p.matcher(input);
		return m.replaceAll(repl);
	}

	/** Helper function
	 * @param types List of types
	 * @return List of types
	 */
	private static LinkedList<String> params (String types)
	{
		LinkedList<String> ret = new LinkedList<String>();
		Pattern p = Pattern.compile(">;");
		String[] blah = p.split(types);
		for (int i=0;i<blah.length;i++)
			ret.add(sed(blah[i]+">", ";", ","));
		return ret;
	}	

	/** Helper function
	 * @param returns Return value
	 * @return Return value
	 */
	private static String retval (String returns)
	{
		//System.out.println(returns);
		String ret = returns.substring(returns.indexOf(")")+1, 
			returns.length());
		ret = sed(ret, ";\\)", "");
		return ret;
	}

	/** Check whether a regular expression matches a string
	 * @param str String
	 * @param regexp Regular expression
	 * @return True if matching, false if not.
	 */
	public static boolean matches (String str, String regexp)
	{
		Pattern p = Pattern.compile(regexp);
		Matcher m = p.matcher(str);
		return m.matches();
	}

	/** Convert all fully-qualified names to normal names
	 * @param str Function declaration
	 * @return Modified function declaration
	 */
	public static String unclassify (String str)
	{
		StringBuffer buf = new StringBuffer();
		StringTokenizer tok = new StringTokenizer(str, " ");
		while (tok.hasMoreTokens())
			buf.append(Refl.lastToken(tok.nextToken(), ".")+" ");
		return buf.toString();
	}

	/** Get a list of methods declared in a certain class
	 * @param klass Fully-qualified class name
	 * @return List of methods
	 */
	public static LinkedList<String> getMethods (String klass)
	{
		LinkedList<String> ret = new LinkedList<String>();
		JavaClass c = Repository.lookupClass(klass);
		if (c == null)
			return ret;
		for (Method m : c.getMethods())
		{
			String mt = Refl.firstToken(m.toString(), "[");
			mt = sed(mt, "\\(", " ( ");
			mt = sed(mt, "\\)", " )");
			mt = sed(mt, " arg[0-9]*", "");
			// Extended regular expressions seem to be broken.
			//mt = sed(mt, " .*?,", "");
			mt = unclassify(mt);
			mt = sed(mt, "\\( ", "(");
			mt = sed(mt, " \\) $", ")");
			
			if (!matches(mt, ".*<init>.*"))
				ret.add(mt);
		}
		return ret;
	}

	/** Get a list of parameters a certain function takes
	 * @param klass Fully-qualified class name
	 * @param fid ID of the function
	 * @return List of parameter types
	 */
	public static LinkedList<TypeTree> getParameters (String klass, int fid)
	{
		JavaClass c = Repository.lookupClass(klass);
		if (c == null)
			return new LinkedList<TypeTree>();
		fid++;
		Method[] m = c.getMethods();
		if (fid > m.length-1)
			return new LinkedList<TypeTree>();
		return handleMethod(m[fid]);
	}

	/** Get the return value of a certain function
	 * @param klass Fully-qualified class name
	 * @param fid Function ID
	 * @return Return value type
	 */
	public static TypeTree getRetVal (String klass, int fid)
	{
		JavaClass c = Repository.lookupClass(klass);
		if (c == null)
			return null;
		fid++;
		Method[] m = c.getMethods();
		if (fid > m.length-1)
			return null;
		String sign = retval(getSign(m[fid]));
		//System.out.println(sign);
		try
		{
			return new TypeTree(sign);
		}
		catch (ClassNotFoundException e) {}
		return null;
	}

	/** Parse the 'signature' of a method
	 * @param m Method
	 * @return Parsed signature
	 */
	private static String getSign (Method m)
	{
		String sign = "[Signature(";
		String str = m.toString();
		int idx = str.indexOf(sign);
		sign = str.substring(idx+sign.length(), str.length()-1);

		sign = sed(sign, "<L", "<");
		sign = sed(sign, ";>", ">");
		sign = sed(sign, ";L", ";");
		sign = sed(sign, "\\)L", ")");
		sign = sed(sign, "/", ".");
		sign = sed(sign, "\\(L", "");
		
		return sign;
	}

	/** Helper function
	 * @param m Method
	 * @return List of types
	 */
	private static LinkedList<TypeTree> handleMethod (Method m)
	{
		String sign = getSign(m);
		int end = sign.indexOf(")");
		LinkedList<String> list = params(sign.substring(0, 
			(end == -1 ? sign.length() : end)));
		
		LinkedList<TypeTree> ret = new LinkedList<TypeTree>();
		for (String s : list)
			try
			{
				ret.add(new TypeTree(s));
			} catch (ClassNotFoundException e) {}
		return ret;
	}
}
