package ReflWrap;

import java.lang.annotation.*;
import java.lang.reflect.*;
import java.util.*;

/** Wrapper around java.lang.Class
 * Retrieve information about classes at runtime.
 */
public class ClassInfo
{
	/** The underlying class */
	private Class klasse;
	/** Prettified name of the class */
	private String name;
	/** Fully qualified name of the class */
	private String fqcn;
	/** Methods in original order */
	private Method[] m;
	/** Methods in correct order according to the interface definition */
	private Method[] m_ord;

	/** Constructor from a fully-qualified classname.
	 * @param FQCN fully-qualified classname
	 */
	public ClassInfo (String FQCN) throws ClassNotFoundException
	{
		this(Class.forName(Refl.firstToken(FQCN, "<")));	

		if (Refl.hasToken(FQCN, "<"))
		{
			String moo = Refl.lastToken(FQCN, "<");
			StringTokenizer tok = new StringTokenizer(moo.substring(0, 
					moo.length()-1), ",");
		}
	}

	/** Constructor from an available Class object
	 * @param klasse Class object
	 */
	public ClassInfo (Class klasse)
	{
		this.klasse = klasse;
		
		fqcn = klasse.getName();
		if (fqcn.equals("java.lang.Boolean"))
			fqcn = "boolean";
		if (fqcn.equals("java.lang.Byte"))
			fqcn = "byte";
		if (fqcn.equals("java.lang.Character"))
			fqcn = "char";
		if (fqcn.equals("java.lang.Double"))
			fqcn = "double";
		if (fqcn.equals("java.lang.Float"))
			fqcn = "float";
		if (fqcn.equals("java.lang.Integer"))
			fqcn = "int";
		if (fqcn.equals("java.lang.Long"))
			fqcn = "long";
		if (fqcn.equals("java.lang.Short"))
			fqcn = "short";
		if (fqcn.equals("java.lang.Void"))
			fqcn = "void";
		
		if (isArray())
			name = arrayType()+"[]";
		else
		{
			StringTokenizer str = new StringTokenizer(fqcn, ".");
			name = null;
			while (str.hasMoreTokens())
				name = str.nextToken();
		}

		m = null;
		m_ord = orderMethods(klasse.getMethods());
	}

	/** Orders methods according to static FunctionIDs (FIDs)
	 * @param old_m Array of all available methods in this class
	 * @return Sorted array of methods
	 */
	private static Method[] orderMethods (Method[] old_m)
	{
		Method[] ret = new Method[old_m.length];

		for (int i=0;i<ret.length;i++)
			if (old_m[i].isAnnotationPresent(sFID.class))
			{
				short fid = old_m[i].getAnnotation(sFID.class).value();
				ret[fid-1] = old_m[i];
				old_m[i] = null;
			}
		
		for (int i=0;i<ret.length;i++)
			if (old_m[i]!=null)
				for (int j=i;j<ret.length;j++)
					if (ret[j]==null)
					{
						ret[j] = old_m[i];
						old_m[i] = null;
					}

		return ret;
	}

	/** Checks if a certain Annotation is present
	 * @param klass Requested Annotation
	 * @return True if the Annotation is present, false if it is not.
	 */
    @SuppressWarnings("unchecked")
	public <C extends Class<? extends Annotation>> boolean isAnot (C klass)
	{
		return klasse.isAnnotationPresent(klass);
	}

	/** Get the fully-qualified classname of this class
	 * @return Fully-qualified classname
	 */
	public String fqcn ()
	{
		return fqcn;
	}

	/** Get the fully-qualified classname in C++ notation
	 * @return Fully-qualified classname
	 */
	public String fqcncs ()
	{
		StringBuffer buf = new StringBuffer();
		StringTokenizer str = new StringTokenizer(fqcn(), ".");
		while (str.hasMoreTokens())
			buf.append(str.nextToken()+"::");
			
		return buf.toString();
	}

	/** Get the filename of this class
	 * @return Filename
	 */
	public String filename ()
	{
		StringBuffer buf = new StringBuffer();
		StringTokenizer str = new StringTokenizer(fqcn(), ".");
		while (str.hasMoreTokens())
			buf.append(str.nextToken()+(str.hasMoreTokens() ? "/" : ".class"));

		return buf.toString();
	}
	
	/** Get the name of the package this class is part of
	 * @return Package name
	 */
	public String _package ()
	{
		StringBuffer buf = new StringBuffer();
		StringTokenizer tok = new StringTokenizer(fqcn(), ".");
		
		while (tok.countTokens()>1)
			buf.append(tok.nextToken()+((tok.countTokens()>2) ? "." : ""));

		return buf.toString();
	}
	
	/** Get the simple name of the class
	 * @return Class name
	 */
	public String name ()
	{
		return name;
	}

	/** Get the underlying Class object
	 * @return Class object
	 */
	public Class _class ()
	{
		return klasse;
	}

	/** Get the all methods this class provides
	 * @return Array of methods
	 */
	public Method[] methods ()
	{
		return m_ord;
	}

	/** Get only the methods which are declared in this class
	 * @return Array of methods
	 */
	public Method[] meth_here ()
	{
		if (m!=null)
			return m;
		
		int num = 0;
		for (int i=0;i<m_ord.length;i++)
			if (m_ord[i].getDeclaringClass().equals(klasse))
				num++;

		m = new Method[num];
		int j = 0;
		for (int i=0;i<m_ord.length;i++)
			if (m_ord[i].getDeclaringClass().equals(klasse))
			{
				m[j] = m_ord[i];
				j++;
			}

		return m;
	}

	/** Get all constructors for this class
	 * @return Array of constructors
	 */
	public Constructor[] constructors ()
	{
		return klasse.getConstructors();
	}

	/** Get all interfaces which are implemented by this class
	 * @return Array of classes
	 */
	public Class[] interfaces ()
	{
		return klasse.getInterfaces();
	}

	/** Check if a certain interface is implemented by this class
	 * @param klass Fully-qualified classname of the interface
	 * @return True if it is implemented by this class, false if it is not.
	 */
	public boolean implementing (String klass)
	{
		if (sclass()!=null)
		{
			ClassInfo supper = new ClassInfo(sclass());
			if (supper.implementing(klass))
				return true;
		}
		
		for (int i=0;i<interfaces().length;i++)
			if (interfaces()[i].getName().equals(klass))
				return true;

		return false;
	}

	/** Get the name of this class' superclass
	 * @return Name of the superclass
	 */
	public String superclass ()
	{
		return (new ClassInfo(klasse.getSuperclass())).name();
	}

	/** Get the fully-qualified name of this class' superclass
	 * @return Fully-qualified name of the superclass
	 */
	public String fqsuperclass ()
	{
		return klasse.getSuperclass().getName();
	}

	/** Get the class object of the superclass
	 * @return Class object of the superclass
	 */
	public Class sclass ()
	{
		return klasse.getSuperclass();
	}

	/** Get a string representation of this object
	 * @return String
	 */
	public String toString ()
	{
		return klasse.toString();
	}

	/** Get the members of this class
	 * @return Array of fields
	 */
	public Field[] fields ()
	{
		return klasse.getDeclaredFields();
	}

	/** Debugging method which prints all declared methods to the screen
	 * @param klasse Requested class
	 */
	public static void dumpMethods (Class klasse)
	{
		Method[] mm = orderMethods(klasse.getMethods());
		
		for (int i=0;i<mm.length;i++)
			System.out.print(i+". "+new MethodInfo(mm[i]));
	}

	/** Retrieve the base type of an array
	 * @return Base type or null if this class is not an array
	 */
	public Class arrayType ()
	{
		int i=0;
		while (fqcn.charAt(i)=='[') i++;

		if (i>0)
		{
			String type = fqcn.substring(i, fqcn.length());
			if (type.equals("Z"))
				return Boolean.TYPE;
			if (type.equals("B"))
				return Byte.TYPE;
			if (type.equals("C"))
				return Character.TYPE;
			if (type.equals("D"))
				return Double.TYPE;
			if (type.equals("F"))
				return Float.TYPE;
			if (type.equals("I"))
				return Integer.TYPE;
			if (type.equals("J"))
				return Long.TYPE;
			if (type.equals("S"))
				return Short.TYPE;

			try
			{
				if (type.charAt(0)=='L')
					return Class.forName(
						fqcn.substring(i+1, fqcn.length()-1));
			} catch (ClassNotFoundException e) {}
			
			System.err.println("ERROR: "+type+" .");
		}
		
		return null;
	}
		
	/** Retrieve a String of the code to generate a new object of
	 * this class.
	 * @return Code
	 */
	public String clstr ()
	{
		if (fqcn.equals("boolean"))
			return "Boolean.TYPE";
		if (fqcn.equals("byte"))
			return "Byte.TYPE";
		if (fqcn.equals("char"))
			return "Character.TYPE";
		if (fqcn.equals("double"))
			return "Double.TYPE";
		if (fqcn.equals("float"))
			return "Float.TYPE";
		if (fqcn.equals("int"))
			return "Integer.TYPE";
		if (fqcn.equals("long"))
			return "Long.TYPE";
		if (fqcn.equals("short"))
			return "Short.TYPE";
		if (fqcn.equals("void"))
			return "Void.TYPE";

		if (klasse.isArray())
		{
			StringBuffer buf = new StringBuffer();
			buf.append("(new "+arrayType());
			int i=0; 
			while (fqcn.charAt(i)=='[') i++;
			for (int j=0;j<i;j++)
				buf.append("[0]");
			buf.append(").getClass()");
			return buf.toString();
		}

		return fqcn+".class";
	}

	/** Compare ClassInfo objects
	 * @param data ClassInfo object
	 * @return True if equal, false if not.
	 */
	public boolean equals (Object data)
	{
		if (!(data instanceof ClassInfo))
			return false;
		
		if (fqcn()!=((ClassInfo)data).fqcn())
			return false;
		
		return true;
	}

	/** Check whether or not this Class is an array type
	 * @return True if yes, false otherwise.
	 */
	public boolean isArray ()
	{
		return klasse.isArray();
	}

	/** Debugging helper
	 * @return List of methods declared in this class
	 */
	public LinkedList<String> methodList ()
	{
		LinkedList<String> ret = new LinkedList<String>();
		for (Method m : methods())
			if (m.getDeclaringClass() == _class())
				ret.add(new MethodInfo(m).toSimpleString());
		return ret;
	}
}
